searchNodes=[{"type":"module","title":"Vettore","doc":"The Vettore library is designed for fast, in-memory operations on vector (embedding) data. All vectors (embeddings) are stored in a Rust data structure (a HashMap ), accessed via a shared resource (using Rustler’s ResourceArc with a Mutex ). Core operations include: Creating a collection : A named set of embeddings with a fixed dimension and a chosen similarity metric ( &quot;hnsw&quot; , &quot;binary&quot; , &quot;euclidean&quot; , &quot;cosine&quot; , or &quot;dot&quot; ). Inserting an embedding : Add a new embedding (with ID, vector, and optional metadata) to a specific collection. Retrieving embeddings : Fetch all embeddings from a collection or look up a single embedding by its unique ID. Similarity search : Given a query vector, calculate a “score” for every embedding in the collection and return the top‑k results (e.g. the smallest distances or largest similarities). Usage Example db = Vettore . new_db ( ) :ok = Vettore . create_collection ( db , &quot;my_collection&quot; , 3 , &quot;euclidean&quot; ) # Insert an embedding via struct: embedding = % Vettore.Embedding { id : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } :ok = Vettore . insert_embedding ( db , &quot;my_collection&quot; , embedding ) # Retrieve it back: { :ok , returned_emb } = Vettore . get_embedding_by_id ( db , &quot;my_collection&quot; , &quot;my_id&quot; ) IO . inspect ( returned_emb . vector , label : &quot;Retrieved vector&quot; ) # Perform a similarity search: { :ok , top_results } = Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.5 , 1.5 , 1.5 ] , 2 ) IO . inspect ( top_results , label : &quot;Top K search results&quot; )","ref":"Vettore.html"},{"type":"function","title":"Vettore.create_collection/5","doc":"Creates a new collection in the database with the given name , dimension , and distance metric. db is the database resource (created with new_db/0 ). name is the name of the collection. dimension is the number of dimensions in the vector. distance can be one of: &quot;euclidean&quot; , &quot;cosine&quot; , &quot;dot&quot; , &quot;hnsw&quot; , or &quot;binary&quot; . Returns {:ok, name} on success, or {:error, reason} if the collection already exists or if the distance is invalid. Examples = Vettore.create_collection(db, &quot;my_collection&quot;, 3, &quot;euclidean&quot;)","ref":"Vettore.html#create_collection/5"},{"type":"function","title":"Vettore.delete_collection/2","doc":"Deletes a collection by its name . db is the database resource (created with new_db/0 ). name is the name of the collection. Returns {:ok, name} if the collection was found and deleted, or {:error, reason} otherwise. Examples { :ok , &quot;my_collection&quot; } = Vettore . delete_collection ( db , &quot;my_collection&quot; )","ref":"Vettore.html#delete_collection/2"},{"type":"function","title":"Vettore.delete_embedding_by_id/3","doc":"Deletes a single embedding by its id . db is the database resource (created with new_db/0 ). collection is the name of the collection. id is the ID of the embedding. Returns {:ok, id} if the embedding was found and deleted, or {:error, reason} otherwise. Examples Vettore . delete_embedding_by_id ( db , &quot;my_collection&quot; , &quot;my_id&quot; ) # =&gt; {:ok, &quot;my_id&quot;}","ref":"Vettore.html#delete_embedding_by_id/3"},{"type":"function","title":"Vettore.get_embedding_by_id/3","doc":"Looks up a single embedding by its id , within the given collection . If found, returns {:ok, %Vettore.Embedding{}} . If not found, returns {:error, reason} . Examples Vettore . get_embedding_by_id ( db , &quot;my_collection&quot; , &quot;my_id&quot; ) # =&gt; {:ok, %Vettore.Embedding{id: &quot;my_id&quot;, vector: [1.0, 2.0, 3.0], metadata: %{foo: &quot;bar&quot;}}}","ref":"Vettore.html#get_embedding_by_id/3"},{"type":"function","title":"Vettore.get_embeddings/2","doc":"Returns all embeddings from a given collection. Each embedding is returned as (id, vector, metadata) in a list—if you want to convert each item into a %Vettore.Embedding{} , you can do so manually or provide a helper function. Examples Vettore . get_embeddings ( db , &quot;my_collection&quot; ) # =&gt; {:ok, [ # {&quot;emb1&quot;, [1.0, 2.0, 3.0], %{&quot;info&quot; =&gt; &quot;test&quot;}}, # {&quot;emb2&quot;, [3.14, 2.71, 1.62], nil}, # ]}","ref":"Vettore.html#get_embeddings/2"},{"type":"function","title":"Vettore.insert_embedding/3","doc":"Inserts a single embedding (as a %Vettore.Embedding{} struct) into a specified collection . If the collection doesn't exist, you'll get {:error, &quot;Collection '...' not found&quot;} . If another embedding with the same :id is already in the collection, you’ll get an error. Also, if the :vector length does not match the collection's configured dimension, you’ll get a dimension mismatch error. Examples embedding = % Vettore.Embedding { id : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ foo : &quot;bar&quot; } } { :ok , &quot;my_id&quot; } = Vettore . insert_embedding ( db , &quot;my_collection&quot; , embedding )","ref":"Vettore.html#insert_embedding/3"},{"type":"function","title":"Vettore.insert_embeddings/3","doc":"Batch-inserts a list of embeddings (each a %Vettore.Embedding{} ) into the specified collection . It returns {:ok, [&quot;id1&quot;, &quot;id2&quot;, ...]} if all embeddings inserted successfully. If any embedding fails (e.g., dimension mismatch, ID conflict, or missing collection), the function returns {:error, reason} and stops immediately (does not insert the rest). Examples embs = [ % Vettore.Embedding { id : &quot;e1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : nil } , % Vettore.Embedding { id : &quot;e2&quot; , vector : [ 4.5 , 6.7 , 8.9 ] , metadata : %{ &quot;info&quot; =&gt; &quot;test&quot; } } ] { :ok , [ &quot;e1&quot; , &quot;e2&quot; ] } = Vettore . insert_embeddings ( db , &quot;my_collection&quot; , embs )","ref":"Vettore.html#insert_embeddings/3"},{"type":"function","title":"Vettore.mmr_rerank/4","doc":"Re-rank a list of {id, score} results using Maximal Marginal Relevance (MMR). Given a database resource db , a collection name, and an initial_results list of {id, score} tuples (usually obtained from similarity_search/4 ), this function applies an MMR formula to select up to :limit items that maximize both relevance (the score ) and diversity among the selected items. The alpha parameter (0.0 to 1.0) balances relevance vs. redundancy: alpha close to 1.0 → prioritizes the raw score (similarity to query). alpha close to 0.0 → heavily penalizes items similar to already-selected ones, thus promoting diversity. We automatically convert Euclidean or Binary distance into a “higher is better” similarity by negating the distance (i.e. similarity = -distance ). For Cosine, Dot Product, or HNSW approaches, the score is already in a higher-is-better format. Returns {:ok, [{id, mmr_score}, ...]} on success or {:error, reason} if the collection is not found. Examples After calling similarity_search/4 : { :ok , initial_results } = Vettore . similarity_search ( db , &quot;my_collection&quot; , query_vec , limit : 50 ) You can re-rank: { :ok , mmr_list } = Vettore . mmr_rerank ( db , &quot;my_collection&quot; , initial_results , limit : 10 , alpha : 0.7 ) mmr_list then gives a smaller set (up to 10 items) in MMR order, each with a new score ( mmr_score ) that reflects their final MMR weighting.","ref":"Vettore.html#mmr_rerank/4"},{"type":"function","title":"Vettore.new_db/0","doc":"Returns a new database resource (wrapped in a Rustler ResourceArc ). The database resource is a handle for the underlying Rust data structure. Examples is_reference ( db ) # =&gt; true","ref":"Vettore.html#new_db/0"},{"type":"function","title":"Vettore.similarity_search/4","doc":"Similarity search with optional limit (defaults to 10) and optional filter map. performs a similarity (or distance) search in the given collection using the provided query vector, returning the top-k results as a list of {embedding_id, score} tuples. For &quot;euclidean&quot; , lower scores are better (distance). For &quot;cosine&quot; , higher scores are better (dot product). For &quot;dot&quot; , also higher is better. For &quot;binary&quot; , the score is the Hamming distance—lower is more similar. For &quot;hnsw&quot; , an approximate nearest neighbors is used. Examples: Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 2 , filter : %{ &quot;category&quot; =&gt; &quot;test&quot; } ) Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 2 ) Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] ) # =&gt; {:ok, [{&quot;emb1&quot;, 0.0}, {&quot;emb2&quot;, 1.23}]}","ref":"Vettore.html#similarity_search/4"},{"type":"module","title":"Vettore.Embedding","doc":"Represents a single embedding entry for insertion into a collection. Fields :id - A unique string identifier for this embedding (e.g. &quot;emb123&quot;). :vector - A list of floating‑point numbers representing the embedding (e.g. [1.0, 2.0, 3.0] ). :metadata - (Optional) A map with any additional information you want to store (e.g. %{&quot;info&quot; =&gt; &quot;my note&quot;} ).","ref":"Vettore.Embedding.html"},{"type":"type","title":"Vettore.Embedding.t/0","doc":"","ref":"Vettore.Embedding.html#t:t/0"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler","doc":"Vettore is an in-memory vector database implemented in Rust and exposed to Elixir via Rustler . It allows you to create collections of vectors (embeddings), insert new embeddings (with optional metadata), retrieve all embeddings or a specific embedding by its ID, and perform similarity searches using common metrics with CPU-specific instructions for maximum performance. The supported distance metrics are: Euclidean Cosine Dot Product HNSW (Hierarchical Navigable Small World graphs for approximate nearest neighbor search) Binary (Compressed vectors with fast Hamming distance)","ref":"readme.html"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Installation","doc":"def deps do [ { :vettore , &quot;~&gt; 0.1.7&quot; } ] end","ref":"readme.html#installation"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Compile","doc":"mix compile This sets the Rust compiler’s “target-cpu” to native, instructing it to generate code optimized for your machine’s CPU features (SSE, AVX, etc.).","ref":"readme.html#compile"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Overview","doc":"The Vettore library is designed for fast, in-memory operations on vector data. All vectors (embeddings) are stored in a Rust data structure (a HashMap) and accessed via a shared resource (using Rustler’s ResourceArc with a Mutex). The core operations include: Creating a collection – A collection is a set of embeddings with a fixed dimension and a chosen similarity metric: hnsw , binary , euclidean , cosine , or dot . Inserting an embedding – Add a new vector with an identifier and optional metadata to a specific collection. Batch Inserting embeddings – Insert a list of embeddings (batch) into a collection in one call. Retrieving embeddings – Fetch all embeddings from a collection or look up a single embedding by its unique ID. Similarity search – Given a query vector, calculate a “score” (distance or similarity) for every embedding in the collection and return the top‑k results.","ref":"readme.html#overview"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Under the Hood","doc":"Data Structures CacheDB The main in‑memory database is defined as a CacheDB struct. Internally, it stores collections in a Rust HashMap&lt;String, Collection&gt; , mapping collection names to a Collection struct. Collection Each collection is a struct with: dimension: usize – The fixed length of every vector in this collection. distance: Distance – The similarity metric used (e.g., Euclidean, Cosine, DotProduct, HNSW, Binary). embeddings: Vec&lt;Embedding&gt; – A vector containing all embeddings. hnsw_index: Option&lt;HnswIndexWrapper&gt; – An optional HNSW index for approximate nearest neighbor search. Embedding An embedding is represented by: id: String – A unique identifier. vector: Vec&lt;f32&gt; – The actual vector values. metadata: Option&lt;HashMap&lt;String, String&gt;&gt; – Optional additional information. binary: Option&lt;Vec&lt;u64&gt;&gt; – For the binary distance metric , a compressed signature. DBResource The CacheDB is wrapped in a DBResource (and stored as a Rustler ResourceArc ). This allows Elixir to hold a reference to the database across NIF calls, and it is guarded by a Mutex to ensure thread safety.","ref":"readme.html#under-the-hood"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Public API / NIF Functions","doc":"All core functions are accessible in Elixir via Vettore.* calls. Their return values (on success) now include more information. 1. new_db/0 Return: a DB resource (a reference to the underlying Rust CacheDB ). Creates a new database resource to be passed to all subsequent calls. Example: db = Vettore . new_db ( ) 2. create_collection/5 Signature: create_collection ( db , name , dimension , distance , opts \\\\ [ ] ) Return: {:ok, collection_name} or {:error, reason} Creates a new collection in the database with a specified dimension and distance metric. The optional opts keyword list may include: keep_embeddings: false — For collections using &quot;binary&quot; distance, this option instructs the NIF to discard the original float vector after compressing it. Examples: # Create a Euclidean collection (default: keep embeddings) { :ok , &quot;euclidean_coll&quot; } = Vettore . create_collection ( db , &quot;euclidean_coll&quot; , 3 , &quot;euclidean&quot; ) # Create a binary collection that does NOT keep the original float vectors: { :ok , &quot;bin_no_keep&quot; } = Vettore . create_collection ( db , &quot;bin_no_keep&quot; , 3 , &quot;binary&quot; , keep_embeddings : false ) # Create a binary collection that DOES keep the original float vectors: { :ok , &quot;bin_keep&quot; } = Vettore . create_collection ( db , &quot;bin_keep&quot; , 3 , &quot;binary&quot; , keep_embeddings : true ) 3. delete_collection/2 Signature: delete_collection ( db , name ) Return: {:ok, collection_name} or {:error, reason} Deletes an existing collection (by name). Example: { :ok , &quot;euclidean_coll&quot; } = Vettore . delete_collection ( db , &quot;euclidean_coll&quot; ) 4. insert_embedding/3 Signature: insert_embedding ( db , collection_name , embedding_struct ) Return: {:ok, embedding_id} or {:error, reason} Inserts a single embedding into a collection. The embedding_struct must be a %Vettore.Embedding{} struct that includes an id , a vector (list of floats), and optional metadata. Example: embedding = % Vettore.Embedding { id : &quot;emb1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;example&quot; } } { :ok , &quot;emb1&quot; } = Vettore . insert_embedding ( db , &quot;euclidean_coll&quot; , embedding ) 5. insert_embeddings/3 Signature: insert_embeddings ( db , collection_name , [ embedding_structs ] ) Return: {:ok, list_of_inserted_ids} or {:error, reason} Batch-inserts a list of embeddings in one call. If any embedding fails (e.g. due to a dimension mismatch or duplicate ID), the function returns an error immediately and does not insert the remaining embeddings. Example: embeddings = [ % Vettore.Embedding { id : &quot;emb1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;first&quot; } } , % Vettore.Embedding { id : &quot;emb2&quot; , vector : [ 2.0 , 3.0 , 4.0 ] , metadata : nil } ] { :ok , [ &quot;emb1&quot; , &quot;emb2&quot; ] } = Vettore . insert_embeddings ( db , &quot;euclidean_coll&quot; , embeddings ) 6. get_embeddings/2 Signature: get_embeddings ( db , collection_name ) Return: {:ok, list_of({id, vector, metadata})} or {:error, reason} Retrieves all embeddings from the specified collection. Each embedding is returned as a tuple: {id, vector, metadata} . Example: { :ok , embeddings } = Vettore . get_embeddings ( db , &quot;euclidean_coll&quot; ) # Example output: [{&quot;emb1&quot;, [1.0, 2.0, 3.0], %{&quot;note&quot; =&gt; &quot;example&quot;}}, ...] 7. get_embedding_by_id/3 Signature: get_embedding_by_id ( db , collection_name , id ) Return: {:ok, %Vettore.Embedding{}} or {:error, reason} Looks up a single embedding by its unique ID within a collection. Example: { :ok , embedding } = Vettore . get_embedding_by_id ( db , &quot;euclidean_coll&quot; , &quot;emb1&quot; ) # embedding is %Vettore.Embedding{id: &quot;emb1&quot;, vector: [1.0, 2.0, 3.0], metadata: %{&quot;note&quot; =&gt; &quot;example&quot;}} 8. similarity_search/4 Signature: similarity_search ( db , collection_name , query_vector , opts \\\\ [ ] ) Return: {:ok, list_of({id, score})} or {:error, reason} Performs a similarity or distance search using the given query vector, returning the top‑k results. Optional parameters: limit: k — The number of top results to return (default is 10). filter: %{} — A metadata filter map (only supported for non‑HNSW collections). Examples: # Basic similarity search (Euclidean: lower distances are better) { :ok , results } = Vettore . similarity_search ( db , &quot;euclidean_coll&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 2 ) # Similarity search with metadata filtering { :ok , filtered_results } = Vettore . similarity_search ( db , &quot;euclidean_coll&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 2 , filter : %{ &quot;category&quot; =&gt; &quot;special&quot; } ) 9. mmr_rerank/4 Signature: mmr_rerank ( db , collection_name , initial_results , opts \\\\ [ ] ) Return: {:ok, list_of({id, mmr_score})} or {:error, reason} Re-ranks a list of {id, score} pairs (e.g., from a previous similarity_search/4 call) using Maximal Marginal Relevance (MMR) . This helps select up to :limit items that are both highly relevant (based on their initial score) and also sufficiently distinct from each other. Optional parameters: limit: k — How many items to keep in the final MMR‑ranked list (default 10). alpha: float — MMR balancing factor in [0..1] . Closer to 1.0 places more emphasis on each candidate’s original score; closer to 0.0 emphasizes diversity. Examples: # 1) Perform a similarity search first. { :ok , initial } = Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 50 ) # &#39;initial&#39; is a list of {id, score} tuples. # 2) Apply MMR re‑ranking on those results: { :ok , mmr_list } = Vettore . mmr_rerank ( db , &quot;my_collection&quot; , initial , limit : 10 , alpha : 0.7 ) # &#39;mmr_list&#39; is now a top-10 list of {id, mmr_score} tuples. 10. delete_embedding_by_id/3 Signature: delete_embedding_by_id ( db , collection_name , id ) Return: {:ok, id} or {:error, reason} Deleting an embedding by its id is as simple as: Examples: { :ok , &quot;my_id&quot; } = Vettore . delete_embedding_by_id ( db , &quot;my_collection&quot; , &quot;my_id&quot; ) # =&gt; {:ok, &quot;my_id&quot;}","ref":"readme.html#public-api-nif-functions"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Complete Usage Example","doc":"Below is an example demonstrating the entire workflow: # Create a new DB resource db = Vettore . new_db ( ) # Create a Euclidean collection { :ok , &quot;euclidean_coll&quot; } = Vettore . create_collection ( db , &quot;euclidean_coll&quot; , 3 , &quot;euclidean&quot; ) # Insert an embedding into the Euclidean collection embedding = % Vettore.Embedding { id : &quot;emb1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;example&quot; } } { :ok , &quot;emb1&quot; } = Vettore . insert_embedding ( db , &quot;euclidean_coll&quot; , embedding ) # Retrieve the embedding by its ID { :ok , emb } = Vettore . get_embedding_by_id ( db , &quot;euclidean_coll&quot; , &quot;emb1&quot; ) IO . inspect ( emb , label : &quot;Retrieved Embedding&quot; ) # Perform a similarity search { :ok , results } = Vettore . similarity_search ( db , &quot;euclidean_coll&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 1 ) IO . inspect ( results , label : &quot;Search Results&quot; ) # Create a binary collection without keeping the raw embeddings { :ok , &quot;bin_no_keep&quot; } = Vettore . create_collection ( db , &quot;bin_no_keep&quot; , 3 , &quot;binary&quot; , keep_embeddings : false ) # Insert an embedding into the binary collection; raw vector will be cleared. { :ok , &quot;nokey1&quot; } = Vettore . insert_embedding ( db , &quot;bin_no_keep&quot; , % Vettore.Embedding { id : &quot;nokey1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : nil } ) { :ok , no_keep_embs } = Vettore . get_embeddings ( db , &quot;bin_no_keep&quot; ) IO . inspect ( no_keep_embs , label : &quot;Binary Collection (no keep)&quot; ) # Create a binary collection that keeps the raw vectors { :ok , &quot;bin_keep&quot; } = Vettore . create_collection ( db , &quot;bin_keep&quot; , 3 , &quot;binary&quot; , keep_embeddings : true ) { :ok , &quot;key1&quot; } = Vettore . insert_embedding ( db , &quot;bin_keep&quot; , % Vettore.Embedding { id : &quot;key1&quot; , vector : [ 9.9 , 8.8 , 7.7 ] , metadata : %{ &quot;foo&quot; =&gt; &quot;bar&quot; } } ) { :ok , keep_embs } = Vettore . get_embeddings ( db , &quot;bin_keep&quot; ) IO . inspect ( keep_embs , label : &quot;Binary Collection (keep)&quot; ) How It Works Storage In-Memory Storage: All data is stored in memory (within the CacheDB struct). No external databases or disk storage is used by default. Collections are stored as key–value pairs in a HashMap. Each collection’s embeddings are stored in a Vec . Inserting Vectors When you call the insert_embedding , insert_embeddings functions: The collection is retrieved from the database. The function checks that the vector’s dimension matches the collection’s defined dimension. It verifies that there isn’t already an embedding with the same ID in the collection. For Cosine distance collections, the vector is normalized before insertion. For Binary distance collections, a binary signature is precomputed and stored (using the sign of each float to produce a bit-packed representation). The new embeddings is then pushed into the collection’s vector of embeddings. Similarity Search The similarity_search function works as follows: It retrieves the target collection and verifies that the query vector’s dimension matches. Depending on the chosen distance metric, it selects an appropriate function: Euclidean: Computes the standard Euclidean distance. Cosine / DotProduct: Computes the dot product (with normalization applied for Cosine). HNSW: Uses a graph-based approach for approximate nearest neighbor search. Binary: Compresses the query vector into a binary signature and computes the Hamming distance between this signature and those of all stored embeddings. For every embedding in the collection, it calculates a “score” between the stored vector (or its compressed representation) and the query. The results are sorted: For Euclidean distance , lower scores (closer to zero) are better. For Cosine/DotProduct , higher scores are considered more similar. For Binary , a lower Hamming distance means the vectors are more similar. Finally, the top‑k results are returned as a list of tuples (embedding_id, score) . Technique/Algorithm Measures Magnitude Sensitive?¹ Scale Invariant?¹ Best Use Cases Pros Cons Euclidean Distance Straight-line distance Yes No Dense data where both magnitude &amp; direction are important Intuitive, widely used, captures magnitude differences Sensitive to scale differences, high dimensionality issues Cosine Similarity Directional similarity (angle) No Yes Text or high-dimensional data where scale invariance is desired Insensitive to magnitude, works well with normalized vectors Ignores magnitude differences Dot Product Combination of direction &amp; magnitude Yes No Applications where both direction &amp; magnitude matter Computationally efficient, captures both aspects Sensitive to vector magnitudes HNSW Indexing Graph-based Approximate Nearest Neighbor Search Dependent on Metric Dependent on Metric Large datasets, real-time search when approximate results are acceptable Sublinear search time , good speed-accuracy trade-off, scalable Approximate results, index build time and memory overhead Binary (Hamming) Fast binary signature comparison using Hamming distance No Yes Applications requiring ultra‑fast approximate searches on large-scale data Extremely fast comparison via bit-level operations, low memory footprint Loses precision due to compression, less suited when exact distances are needed","ref":"readme.html#complete-usage-example"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Batch Insert Example","doc":"Because insert_embeddings/3 now accepts a list of %Vettore.Embedding{} structs and returns a list of IDs, you can do the following: db = Vettore . new_db ( ) { :ok , &quot;my_coll&quot; } = Vettore . create_collection ( db , &quot;my_coll&quot; , 3 , &quot;cosine&quot; ) # Define a list of embeddings batch = [ % Vettore.Embedding { id : &quot;a&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;tag&quot; =&gt; &quot;alpha&quot; } } , % Vettore.Embedding { id : &quot;b&quot; , vector : [ 2.0 , 2.5 , 3.5 ] , metadata : nil } ] # Insert them all in one call case Vettore . insert_embeddings ( db , &quot;my_coll&quot; , batch ) do { :ok , inserted_ids } -&gt; IO . inspect ( inserted_ids , label : &quot;Batch inserted IDs&quot; ) { :error , reason } -&gt; IO . puts ( &quot;Failed to insert batch: \#{ reason } &quot; ) end On success, you might see something like: Batch inserted IDs : [ &quot;a&quot; , &quot;b&quot; ]","ref":"readme.html#batch-insert-example"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Example Usage for Single Insert &amp; Retrieval","doc":"defmodule VettoreExample do alias Vettore.Embedding def demo do # 1) Create a new in-memory DB resource db = Vettore . new_db ( ) # 2) Create a new collection named &quot;my_collection&quot; { :ok , &quot;my_collection&quot; } = Vettore . create_collection ( db , &quot;my_collection&quot; , 3 , &quot;euclidean&quot; ) # 3) Insert a single embedding embed = % Embedding { id : &quot;emb1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;info&quot; =&gt; &quot;test&quot; } } { :ok , &quot;emb1&quot; } = Vettore . insert_embedding ( db , &quot;my_collection&quot; , embed ) # 4) Retrieve all embeddings { :ok , all_embs } = Vettore . get_embeddings ( db , &quot;my_collection&quot; ) IO . inspect ( all_embs , label : &quot;All embeddings in my_collection&quot; ) # 5) Retrieve specific embedding by ID { :ok , % Embedding { id : e_id , vector : e_vec , metadata : e_meta } } = Vettore . get_embedding_by_id ( db , &quot;my_collection&quot; , &quot;emb1&quot; ) IO . inspect ( e_id , label : &quot;Single embedding ID&quot; ) IO . inspect ( e_vec , label : &quot;Single embedding vector&quot; ) IO . inspect ( e_meta , label : &quot;Single embedding metadata&quot; ) # 6) Similarity search { :ok , top_results } = Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 2 , filter : %{ &quot;info&quot; =&gt; &quot;test&quot; } ) IO . inspect ( top_results , label : &quot;Similarity search results&quot; ) end end","ref":"readme.html#example-usage-for-single-insert-retrieval"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Performance Notes","doc":"HNSW can speed up searches significantly for large datasets but comes with higher memory usage for the index. Binary distance uses bit-level compression and Hamming distance for extremely fast approximate similarity checks (especially beneficial for large or high-dimensional vectors, though it trades off some precision). Cosine normalizes vectors once on insertion, so queries and stored embeddings use a straightforward dot product. Dot Product directly multiplies corresponding elements. Euclidean uses a SIMD approach ( wide::f32x4 ) for partial vectorization. You can further optimize by compiling with: RUSTFLAGS = &quot;-C target-cpu=native -C opt-level=3&quot; mix compile","ref":"readme.html#performance-notes"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - Contributing","doc":"Contributions are welcome! Feel free to open issues or submit pull requests on GitHub.","ref":"readme.html#contributing"},{"type":"extras","title":"Vettore: In-Memory Vector Database with Elixir & Rustler - License","doc":"This project is licensed under the Apache License, Version 2.0 .","ref":"readme.html#license"},{"type":"extras","title":"LICENSE","doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright [yyyy] [name of copyright owner] Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html"},{"type":"extras","title":"performance","doc":"","ref":"performance.html"},{"type":"extras","title":"performance - Command to run","doc":"mix run bench/vettore_bench.exs","ref":"performance.html#command-to-run"},{"type":"extras","title":"performance - My PC Results (Intel(R) Core(TM) Ultra 5 125H)","doc":"Operating System: Linux CPU Information: Intel(R) Core(TM) Ultra 5 125H Number of Available Cores: 18 Available memory: 15.29 GB Elixir 1.18.2 Erlang 27.2.2 JIT enabled: true Benchmark suite executing with the following configuration: warmup: 2 s time: 3 s memory time: 0 ns reduction time: 0 ns parallel: 1 inputs: none specified Estimated total run time: 1 min 15 s Benchmarking batch_insert_1000_binary ... Benchmarking batch_insert_1000_cosine ... Benchmarking batch_insert_1000_dot ... Benchmarking batch_insert_1000_euclidean ... Benchmarking batch_insert_1000_hnsw ... Benchmarking similarity_search_binary ... Benchmarking similarity_search_cosine ... Benchmarking similarity_search_dot ... Benchmarking similarity_search_euclidean ... Benchmarking similarity_search_hnsw ... Benchmarking single_insert_binary ... Benchmarking single_insert_cosine ... Benchmarking single_insert_dot ... Benchmarking single_insert_euclidean ... Benchmarking single_insert_hnsw ... Calculating statistics... Formatting results... Name ips average deviation median 99th % single_insert_hnsw 620.45 K 1.61 μs ±1984.02% 1.37 μs 2.82 μs single_insert_dot 585.92 K 1.71 μs ±2093.49% 1.38 μs 2.89 μs single_insert_binary 562.06 K 1.78 μs ±2378.59% 1.36 μs 2.67 μs single_insert_euclidean 534.20 K 1.87 μs ±1710.27% 1.92 μs 3.06 μs single_insert_cosine 503.73 K 1.99 μs ±2263.61% 1.91 μs 2.77 μs similarity_search_binary 16.43 K 60.85 μs ±46.42% 59.76 μs 150.58 μs similarity_search_euclidean 13.91 K 71.88 μs ±48.66% 73.92 μs 139.96 μs similarity_search_dot 13.57 K 73.68 μs ±23.90% 74.43 μs 120.25 μs similarity_search_cosine 11.86 K 84.29 μs ±33.05% 82.12 μs 172.90 μs similarity_search_hnsw 10.25 K 97.57 μs ±41.99% 93.62 μs 198.98 μs batch_insert_1000_hnsw 3.87 K 258.28 μs ±18.43% 262.59 μs 377.67 μs batch_insert_1000_euclidean 3.83 K 260.94 μs ±38.08% 249.53 μs 599.66 μs batch_insert_1000_binary 3.73 K 267.74 μs ±19.36% 266.81 μs 447.58 μs batch_insert_1000_dot 3.59 K 278.87 μs ±35.56% 258.45 μs 650.09 μs batch_insert_1000_cosine 3.53 K 282.92 μs ±34.05% 307.24 μs 586.30 μs Comparison: single_insert_hnsw 620.45 K single_insert_dot 585.92 K - 1.06x slower +0.0950 μs single_insert_binary 562.06 K - 1.10x slower +0.167 μs single_insert_euclidean 534.20 K - 1.16x slower +0.26 μs single_insert_cosine 503.73 K - 1.23x slower +0.37 μs similarity_search_binary 16.43 K - 37.76x slower +59.24 μs similarity_search_euclidean 13.91 K - 44.60x slower +70.27 μs similarity_search_dot 13.57 K - 45.71x slower +72.07 μs similarity_search_cosine 11.86 K - 52.30x slower +82.68 μs similarity_search_hnsw 10.25 K - 60.54x slower +95.96 μs batch_insert_1000_hnsw 3.87 K - 160.25x slower +256.67 μs batch_insert_1000_euclidean 3.83 K - 161.90x slower +259.33 μs batch_insert_1000_binary 3.73 K - 166.12x slower +266.13 μs batch_insert_1000_dot 3.59 K - 173.02x slower +277.26 μs batch_insert_1000_cosine 3.53 K - 175.54x slower +281.31 μs","ref":"performance.html#my-pc-results-intel-r-core-tm-ultra-5-125h"},{"type":"extras","title":"performance - 1. Single Insert Benchmarks","doc":"Single insert across all distance metrics (HNSW, dot, binary, euclidean, cosine) are all in the range of 1.6–2.0 μs . The differences here are minor: HNSW was unexpectedly the quickest for single insert (1.61 μs) Then dot (1.71 μs), binary (1.78 μs), euclidean (1.87 μs), cosine (1.99 μs). Why So Similar? For small dimension (3D) and a small number of embeddings, the overhead of checking dimension + metadata + pushing to an internal Vec is likely dominating. HNSW is also quick for single insert because the data set is tiny, so updating the HNSW graph is relatively cheap.","ref":"performance.html#1-single-insert-benchmarks"},{"type":"extras","title":"performance - 2. Batch Insert (1000 Embeddings)","doc":"Times are around 260–280 μs for all metrics, which is extremely fast to insert 1000 embeddings. Differences are small: Metric Average time HNSW ~258.28 μs Euclidean ~260.94 μs Binary ~267.74 μs Dot ~278.87 μs Cosine ~282.92 μs Again, for 3D vectors and only 1000 embeddings, the overhead is quite low. Surprising to see HNSW top the list, but with so few embeddings, building an incremental small graph is cheap. In real large-scale usage, HNSW might trade off more memory for bigger data sets but scale better in searches.","ref":"performance.html#2-batch-insert-1000-embeddings"},{"type":"extras","title":"performance - 3. Similarity Search","doc":"You tested searching for the top‑k (k=10) out of 1000 embeddings. The results (in microseconds) were: Metric Time Observed IPS Binary ~60.85 μs 16.43 K Euclidean ~71.88 μs 13.91 K Dot ~73.68 μs 13.57 K Cosine ~84.29 μs 11.86 K HNSW ~97.57 μs 10.25 K Observations Binary was the fastest search. Hamming distance is just bitwise XOR + popcount, which is very fast especially for small vectors (3D is basically a few bits). Euclidean/Dot/Cosine are next, all fairly close. HNSW is slowest at this tiny scale . HNSW typically shines with bigger data sets, because it can skip a large portion of the embeddings. At 1000 embeddings, a naive full scan can actually be faster than building/traversing a hierarchical graph. For bigger dimension (e.g., 128D, 512D) or bigger data sets (50k–1M embeddings), you might see HNSW start to outperform direct pairwise scans. Right now, the overhead of index traversal is bigger than the cost of scanning a small dataset.","ref":"performance.html#3-similarity-search"},{"type":"extras","title":"performance - Conclusions","doc":"For tiny data (3D, 1k embeddings), all inserts are under a microseconds scale or a couple hundred microseconds for batch. They’re basically near each other. Search at this scale is also very quick for direct metrics. HNSW is slower because the indexing overhead is not amortized. If you tested larger dimensions (say 128D) or larger data (like 100k or 1M embeddings), you’d likely see bigger differences: HNSW might become significantly faster at searching. Dot vs Euclidean vs Binary might shift in overhead, especially for large vectors. At a small scale, these differences are overshadowed by overheads of function calls, iteration, or index management. Single Insert : ~1–2 μs each, nearly identical across metrics. Batch Insert (1000) : ~250–280 μs, HNSW ironically a bit faster, but all close. Similarity Search : Binary is fastest, then Euclidean, Dot, Cosine, and last HNSW. Reason : With only 1000 embeddings &amp; 3D vectors, a naive linear scan is extremely fast, overshadowing the benefits of advanced structures like HNSW. HNSW typically helps more with larger data sets and higher dimensions.","ref":"performance.html#conclusions"}]